._page_8fb7q_1{position:relative;width:100%;height:calc(100% - 74px);flex:1}._rest_8fb7q_8{display:flex;flex:1}._icons_8fb7q_13{flex-shrink:0}._form_1rpft_1{position:absolute;display:flex;flex-direction:column;z-index:1000;right:1rem;bottom:1rem;padding:1rem;font-size:1rem;width:300px;background:var(--background);border:1px solid var(--background-border);box-shadow:0 2px 40px var(--background-dark);border-radius:3px;color:var(--font)}._header_1rpft_20{display:flex;justify-content:space-between;align-items:center;position:relative}._text_1rpft_27{font-weight:unset;margin-bottom:1rem;color:var(--font-dark)}._form_1rpft_1 input{width:100%;background-color:transparent;border:none;padding:.25rem 0;margin-bottom:1rem;color:var(--font);line-height:1rem;border:none;border-bottom:1px solid var(--brand);font-weight:700;outline:0}._teamsForm_1rpft_47 input:focus{border-bottom:1px solid var(--brand);outline:0;font-weight:700}._teamsForm_1rpft_47 input::placeholder{color:var(--font-dark);font-weight:700}