import{R as s,s as a,j as e,w as t,x as o,y as r,z as n,A as f,B as c}from"./index.bf899a34.js";export{z as transformDataset}from"./index.bf899a34.js";import{d,a as l,B as i,R as p}from"./vendor.b0739682.js";const u=c,x=({prepared:c})=>{const{dataset:x}=d.usePreloadedQuery(u,c),m=l.exports.useContext(s),j=i(a),b=e();return l.exports.useEffect((()=>{b((({reset:s})=>(s(t),s(o),{colorscale:m.state.colorscale,config:m.state.config?r(m.state.config):void 0,dataset:n(x),state:m.state.state})))}),[x,c,m]),j?p.createElement(f,null):null};export{x as default};
